import 'dart:convert';

import 'package:irishman/irishman.dart';
import 'dart:io';

/// A class that stores data about a programming language. The data stored is
/// everything, from metadata, to token data, to AST information, etc.
class Language {
  String name = "";
  String description = "";
  String emoji = "";

  List<String> keywords = [];
  List<LanguageSymbol> symbols = [];
  List<LanguageSpan> spans = [];

  /// The default constructor for a language class.
  Language(
    this.name, {
    this.keywords = const [],
    this.symbols = const [],
    this.spans = const [],
  });

  /// Converts a JSON file into a language class.
  factory Language.fromJSONFile(String path) {
    String fileData = File(path).readAsStringSync();
    return Language.fromJSONString(fileData);
  }

  /// Converts a JSON string into a language class.
  factory Language.fromJSONString(String jsonText) {
    Map<String, dynamic> json = jsonDecode(jsonText);

    return Language.fromJson(json);
  }

  /// Coverts JSON data into a language class.
  Language.fromJson(Map<String, dynamic> json) {
    name = json['name'] as String;
    description = json['description'] as String;
    emoji = json['emoji'] as String;

    if (json['syntax'] != null) {
      if (json['syntax']['tokens'] != null) {
        if (json['syntax']['tokens']['keywords'] != null) {
          keywords = List.from(json['syntax']['tokens']['keywords']);
        }

        if (json['syntax']['tokens']['symbols'] != null) {
          List symbolList = json['syntax']['tokens']['symbols'] as List;
          symbols = symbolList
              .map(
                (e) => LanguageSymbol.fromJson(e),
              )
              .toList();
        }

        if (json['syntax']['tokens']['spans'] != null) {
          List spanList = json['syntax']['tokens']['spans'] as List;
          spans = spanList
              .map(
                (e) => LanguageSpan.fromJson(e),
              )
              .toList();
        }
      }
    }
  }

  /// Gets a tokenizer from a language definition.
  Tokenizer getTokenizer() {
    // If we don't presort keywords by length, they will falsely trigger
    // For example: `do` could trigger when it should be `doodle`
    List<String> sortedKeywords = [...keywords];
    sortedKeywords.sort((a, b) => b.length.compareTo(a.length));

    // Same goes for the symbols
    List<LanguageSymbol> sortedSymbols = [...symbols];
    sortedSymbols
        .sort((a, b) => b.activator.length.compareTo(a.activator.length));

    // If we sort by length of initiator, all of the initiator = "" spans
    // or (in other words) those that have start properties, will be first.
    List<LanguageSpan> sortedSpans = [...spans];
    sortedSpans
        .sort((a, b) => b.initiator.length.compareTo(a.initiator.length));

    // Return the finalized tokenizer.
    return Tokenizer(
      keywords: [
        ...sortedKeywords.map(
          (e) => Keyword(e),
        ),
      ],
      symbols: [
        ...sortedSymbols.map(
          (e) => Symbol(e.activator, e.nickname),
        ),
      ],
      spans: [
        ...sortedSpans.map(
          (e) => Span(
            e.nickname,
            start: e.start,
            end: e.end,
            initiator: RegExp(e.initiator),
            conditional: RegExp(e.conditional),
          ),
        )
      ],
    );
  }
}

/// A temporary storage form for Symbols allowing for the system to
/// take in JSON input.
class LanguageSymbol {
  /// The name of the token.
  String nickname = "";

  /// A sequence of characters that activates a symbol token.
  String activator = "";

  /// The default constructor for a language symbol.
  LanguageSymbol(this.nickname, this.activator);

  /// Converts JSON data into a language symbol.
  LanguageSymbol.fromJson(Map<String, dynamic> json) {
    nickname = json['name'] as String;
    activator = json['activator'] as String;
  }
}

/// A temporary storage form for Spans allowing for the system to
/// take in JSON input.
class LanguageSpan {
  /// The name of the token.
  String nickname = "";

  /// A sequence of characters that starts a span token.
  String start = "";

  /// A sequence of characters that ends a span token.
  String end = "";

  /// A sequence of characters that starts a span token.
  String initiator = "";

  /// A sequence of characters that ends a span token.
  String conditional = "";

  /// The default constructor for a language span.
  LanguageSpan(
    this.nickname, {
    required this.start,
    required this.end,
    required this.initiator,
    required this.conditional,
  });

  /// Converts JSON data into a language span.
  LanguageSpan.fromJson(Map<String, dynamic> json) {
    nickname = json['name'] as String;

    if (json['start'] != null && json['end'] != null) {
      start = json['start'] as String;
      end = json['end'] as String;
      initiator = "";
      conditional = "";
    } else if (json['initiator'] != null && json['conditional'] != null) {
      start = "";
      end = "";
      initiator = json['initiator'] as String;
      conditional = json['conditional'] as String;
    } else {
      start = "";
      end = "";
      initiator = "";
      conditional = "";
    }
  }
}
