/// A class that converts standard text into a string of tokens.
class Tokenizer {
  final List<Keyword> keywords;
  final List<Symbol> symbols;
  final List<Span> spans;

  /// The default constructor for the tokenizer.
  const Tokenizer({
    required this.keywords,
    required this.symbols,
    required this.spans,
  });

  /// Tokenizes a string based on the language rules given.
  List<TokenData> tokenizeString(String input) {
    /// Initialize output
    List<TokenData> output = [];

    // Loop through the characters
    int i = 0;
    while (i < input.length) {
      // Loop through each keyword
      for (var keyword in keywords) {
        // Create incrementor and matching bool
        int j = 0;
        bool matching = true;

        // Loop through each character
        do {
          if (i + j >= input.length || j >= keyword.activator.length) {
            matching = false;
            break;
          }
          if (input[i + j] != keyword.activator[j]) matching = false;
          j++;
        } while (j < keyword.activator.length && matching);

        if (matching) {
          output.add(TokenData(
              keyword.getTokenName(), i, i + j, input.substring(i, i + j)));
          i += j - 1;
        }
      }

      // Loop through each symbol
      for (var symbol in symbols) {
        // Create incrementor and matching bool
        int j = 0;
        bool matching = true;

        // Loop through each character
        do {
          if (i + j >= input.length || j >= symbol.activator.length) {
            matching = false;
            break;
          }
          if (input[i + j] != symbol.activator[j]) matching = false;
          j++;
        } while (j < symbol.activator.length && matching);

        if (matching) {
          output.add(TokenData(
              symbol.getTokenName(), i, i + j, input.substring(i, i + j)));
          i += j - 1;
        }
      }

      // Loop through each span and see if the text is matching
      for (var span in spans) {
        if (span.usesStartAndEnd()) {
          // Create incrementor and matching bool
          int j = 0;
          bool matching = true;

          // Loop through each character
          do {
            if (i + j >= input.length || j >= span.start.length) {
              matching = false;
              break;
            }
            if (input[i + j] != span.start[j]) matching = false;
            j++;
          } while (j < span.start.length && matching);

          if (matching) {
            i += j - 1;
            j = 0;
            int endMatchAmount = 0;

            // Now its time to find the end of the span
            do {
              if (input[i + j] == span.end[endMatchAmount]) {
                endMatchAmount++;
              } else {
                endMatchAmount = 0;
              }
              j++;
            } while (i + j < input.length && endMatchAmount < span.end.length);

            if (endMatchAmount == span.end.length) {
              output.add(
                TokenData(
                  span.getTokenName(),
                  (i - 1) + span.start.length,
                  (i + j) - span.end.length,
                  input.substring(
                    (i - 1) + span.start.length,
                    (i + j) - span.end.length,
                  ),
                ),
              );
              i += j - 1;
            }
          }
        } else {}
      }

      i++;
    }

    return output;
  }

  /// Gets all the token names stored within the tokenizer.
  /// The list of token names is returned in alphabetically order.
  List<String> getAllTokenNames() {
    List<String> output = [];

    for (var e in keywords) {
      output.add(e.getTokenName());
    }

    for (var e in symbols) {
      output.add(e.getTokenName());
    }

    for (var e in spans) {
      output.add(e.getTokenName());
    }

    /// Sorts the output tokens alphabetically.
    output.sort((a, b) => a.compareTo(b));

    return output;
  }
}

/// The data stored in each token.
class TokenData {
  /// The token identifier.
  final String tokenName;

  /// The start of the token.
  final int start;

  /// The end of the token.
  final int end;

  /// The data contained inside the token.
  final String data;

  /// Constructor for the TokenData class.
  const TokenData(this.tokenName, this.start, this.end, this.data);
}

/// A type of token used for very intricate text data such as comments,
/// identifiers, numbers, etc.
class Span {
  /// The unofficial name of the token.
  final String nickname;

  /// The optional start of a Span. Must be paired with an end.
  String start;

  /// The optional end of a Span. Must be paired with a start.
  String end;

  /// The optional initiator of a Span. Must be paired with a conditional.
  Pattern initiator;

  /// The optional conditional of a Span. Must be paired with a initiator.
  Pattern conditional;

  /// Constructor for the Span class.
  Span(
    this.nickname, {
    this.start = "",
    this.end = "",
    this.conditional = "",
    this.initiator = "",
  });

  /// Gets if the span depends on start/end or initiator/conditional.
  bool usesStartAndEnd() {
    return start != "" && end != "";
  }

  /// Gets the official name of the token.
  ///
  /// ```dart
  /// Keyword(">=", "Greater than or equals").getTokenName() == "GREATER_THAN_OR_EQUALS"
  /// ```
  String getTokenName() {
    // All token names should be uppercase and use underscores as spaces
    return nickname.toUpperCase().replaceAll(RegExp("\\s"), "_");
  }
}

/// A token type used for symbols and intricate sequences such as `>=`, `||`,
/// `%`, `+`, etc.
class Symbol {
  /// The trigger that creates the Symbol.
  final String activator;

  /// The unofficial name of the token.
  final String nickname;

  /// Constructor for the Symbol class.
  const Symbol(this.activator, this.nickname);

  /// Gets the official name of the token.
  ///
  /// ```dart
  /// Keyword(">=", "Greater than or equals").getTokenName() == "GREATER_THAN_OR_EQUALS"
  /// ```
  String getTokenName() {
    // All token names should be uppercase and use underscores as spaces
    return nickname.toUpperCase().replaceAll(RegExp("\\s"), "_");
  }
}

/// A type of token used for keywords in programming languages such as `class`,
/// `var`, `if`, etc.
class Keyword {
  /// The trigger that creates the keyword.
  final String activator;

  /// Constructor for the Keyword class.
  const Keyword(this.activator);

  /// Gets the official name of the token.
  ///
  /// ```dart
  /// Keyword("if").getTokenName() == "IF_KEYWORD"
  /// ```
  String getTokenName() {
    // All token names should be uppercase and use underscores as spaces
    return "${activator.toUpperCase()}_KEYWORD";
  }
}
